/*
    Copyright (C) 2008-2009 Stefan Haller

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <config.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "../misc/general.h"
#include "../misc/xml.h"
#include "../misc/module.h"
#include "../dbus/dbus-server.h"
#include "../modules/gwallpapers_module.h"

#define MAX_FILTER 7
const char * FILTER [MAX_FILTER] =
		{"png",
		 "jpg",
		 "jpeg",
		 "bmp",
		 "svg",
		 "gif",
		 "tif"
		};

GPtrArray * modules;

gboolean changebyinterval;
guint interval;
gboolean changeeverylaunch;
gboolean use_cache;

gboolean HUP = FALSE;
gboolean quit = FALSE;

GPtrArray * wallpapers = NULL;

GMainLoop * main_loop = NULL;

GTimer * timer = NULL;

#define HISTORY_MAX 10
guint history[HISTORY_MAX];
gint history_current;
gint history_size;

struct file_node * files;

void signals_handler(int sig)
{
	g_message("Received signal \"%s.\"", g_strsignal(sig));
	if (sig == SIGHUP)
	{
		HUP = TRUE;
	}
	if (main_loop != NULL)
	{
		g_main_loop_quit(main_loop);
	}
	else
	{
		quit = TRUE;
	}
}

void config_signals()
{
	struct sigaction sig;
	sig.sa_handler = signals_handler;
	sig.sa_flags = SA_RESTART;
	sigemptyset(&sig.sa_mask);
	
	sigaction(SIGHUP, &sig, NULL);
	sigaction(SIGTERM, &sig, NULL);
	sigaction(SIGINT, &sig, NULL);
	sigaction(SIGQUIT, &sig, NULL);
}

void read_modules(struct module_node * m_node)
{
	while (m_node != NULL)
	{
		gchar * abs_fname = g_build_filename(PKGLIBDIR, m_node->name, NULL);
		/* mustn't be unloaded, because we use gconf (and glib) with static
		   variables in the module! */
		struct library_entry * lib_ent = load_module(abs_fname);
		if (lib_ent != NULL)
		{
			g_ptr_array_add(modules, lib_ent);
		}
		m_node = m_node->next;
		g_free(abs_fname);
	}
}

gboolean check_extension(const gchar * filename)
{
	/* get last . */
	gchar * found = g_strrstr(filename, ".\0");
	if (found == NULL)
	{
		return FALSE;
	}
	
	/* Is dot last character in string? */
	if (found == filename + strlen(filename))
	{
		return FALSE;
	}
	/* remove dot */
	found += 1;
	
	gboolean result = FALSE;
	int i;
	
	for (i = 0; i < MAX_FILTER; i++)
	{
		if (g_strcasecmp(found, FILTER[i]) == 0)
		{
			result = TRUE;
			break;	
		}
	}
	
	return result;
}

void add_wallpaper(GPtrArray * array, const gchar * filename)
{
	/* check file extension */
	if (! check_extension(filename))
		return;	
	
	/* quick and dirty: Can we read the file? */
	int fd = open(filename, O_RDONLY);
	if (fd > -1)
	{
		/* file opened - we can read :) */
		gchar * filename_dup = g_strdup(filename);
		g_ptr_array_add(array, filename_dup);
	}
	close(fd);
}

void clear_wallpapers(GPtrArray * array)
{
	guint i;
	for (i = 0; i < array->len; i++)
	{
		gchar * filename = (gchar *)g_ptr_array_index(array, i);
		g_free(filename);
	}
	g_ptr_array_free(array, TRUE);
}

void change_wallpaper(guint nr)
{
	if (nr >= wallpapers->len)
	{
		return;
	}
	
	g_timer_start(timer);
	gchar * file = g_strdup((const gchar *)g_ptr_array_index(wallpapers, nr));
	
	#ifdef DEBUG
	g_debug("change_wallpaper(%s)", file);
	#endif
	
	
	guint i;
	for (i = 0; i < modules->len; i++)
	{
		struct library_entry * lib = g_ptr_array_index(modules, i);
		if (lib->change_wallpaper != NULL)
		{
			(*lib->change_wallpaper)(file);
		}
	}
	
	#ifdef USE_DBUS
	emit_changed(file);
	#endif
	
	g_free(file);
}

void next_wallpaper()
{
	guint i;
	if (history_current == HISTORY_MAX -1)
	{
		for (i = 0; i < HISTORY_MAX - 1; i++)
		{
			history[i] = history[i+1];
		}
		
		history[HISTORY_MAX - 1] = rand() % wallpapers->len;
		change_wallpaper(history[HISTORY_MAX - 1]);
	}
	else
	{
		if (history_current >= history_size - 1)
		{
			history_size++;
			history_current = history_size - 1;
			history[history_current] = rand() % wallpapers->len;
			change_wallpaper(history[history_current]);
		}
		else
		{
			change_wallpaper(history[++history_current]);
		}
	}
}

void previous_wallpaper()
{
	guint i;
	if (history_current == 0)
	{
		for (i = HISTORY_MAX; i > 0; i--)
		{
			history[i] = history[i-1];
		}
		
		history[0] = rand() % wallpapers->len;
		change_wallpaper(history[0]);
		
		if (history_size != HISTORY_MAX)
			history_size++;
	}
	else
	{
		change_wallpaper(history[--history_current]);
	}
}

gboolean next_wallpaper_timer(gpointer data)
{
	/* 5 seconds tolerance */
	if (g_timer_elapsed(timer, NULL) > (interval * 60) - 5)
	{
		next_wallpaper();
	}
	return TRUE;
}

gboolean change_launch_timer(gpointer data)
{
	next_wallpaper();
	
	#ifndef USE_DBUS
	if (! changebyinterval)
	{
		g_main_loop_quit(main_loop);
	}
	#endif
	
	return FALSE;
}

void browse_directory(GPtrArray * array, const gchar * dirname, gboolean subfolders)
{
	GDir * dir = g_dir_open(dirname, 0, NULL);
	if (dir != NULL)
	{
		const gchar * filename;
		while ((filename = g_dir_read_name(dir)) != NULL)
		{
			gchar * absolute_filename = g_build_filename(dirname,
			                                             filename,
			                                             NULL);
			struct stat buf;
			if (/*g_*/stat(absolute_filename, &buf) == 0)
			{
				if (S_ISDIR(buf.st_mode))
				{
					if (subfolders)
					{
						browse_directory(array, absolute_filename, subfolders);
					}
				}
				else
				{
					add_wallpaper(array, absolute_filename);
				}
			}
			else
			{
				/*
				TODO: Error handling
				*/
				error("stat");
			}
			g_free(absolute_filename);
		}
		g_dir_close(dir);
	}
}

void * scan_files()
{
	struct file_node * node = files;
	while (node != NULL)
	{
		if (node->is_dir)
		{
			browse_directory(wallpapers, node->filename, node->use_subfolders);
		}
		else
		{
			add_wallpaper(wallpapers, node->filename);
		}
		node = node->next;
	}
	
	return NULL;
}

/* create / check pid lock in ~/.config/gwallpapers */
gboolean check_pid_file(char ** fname)
{
	*fname = NULL;
	gchar * app_dir = get_app_dir();
	
	if (app_dir == NULL)
	{
		g_critical("Could not get application directory. Maybe environment " \
		           "variable $HOME isn't set.\n");
		return FALSE;
	}	
	
	gchar * filename = g_build_filename(app_dir, PIDLOCKFILE, NULL);
	g_free(app_dir);
	*fname = filename;
	
	int pidlock = open(filename, O_WRONLY | O_CREAT | O_EXCL, S_IRUSR |
	                             S_IWUSR);
	
	if (pidlock == -1)
	{
		if (errno == EEXIST)
		{
			/* file exists already, another gwallpapers-daemon must be running
			   or a gwallpapers-daemon did not shutdown correctly             */
			/* get pid from pid lock file                                     */
			gboolean quit = FALSE;
			gboolean valid_pid = FALSE;
			gchar * pid = g_strdup("");
			pidlock = open(filename, O_RDONLY);
			if (pidlock != -1)
			{
				gchar c;
				while (read(pidlock, &c, 1) > 0)
				{
					valid_pid = g_ascii_isdigit(c);
					if (! valid_pid)
					{
						break;
					}
					
					gchar * tmp = pid;
					gchar tmp2[2] = {'\0', '\0'};
					tmp2[0] = c;
					pid = g_strconcat(tmp, tmp2, NULL);
					g_free(tmp);
				}
				
				close(pidlock);
			}
			
			if (valid_pid)
			{
				/* check if process with this pid is running */
				gboolean running = TRUE;
				gchar * procfilename = g_build_filename("/proc/", pid, NULL);
				struct stat buf;
				if (stat(procfilename, &buf) == -1)
				{
					running = FALSE;
				}
				g_free(procfilename);
				
				if (running)
				{
					quit = TRUE;
					g_message("%s already exists. Another gWallpapers-daemon " \
					          "is running (PID=%s)! Terminating...", filename, pid);
				}
				else
				{
					g_message("%s already exists, but no process with PID %s "\
					          "is running. However daemon will continue "\
					          "starting.", filename, pid);
				}
			}
			else
			{
				/* no valid pid in pidfile
				   nevertheless start daemon */
				g_message("%s already exists, but couldn't read valid " \
				          "PID. However daemon will continue starting.",
				          filename);
			}
			
			g_free(pid);
			
			if (quit)
			{
				return FALSE;
			}
		}
		else
		{
			g_warning("Could not open pid lock file readably.");
			error("open");
			return FALSE;
		}
	}
	else
	{
		close(pidlock);
	}
	
	FILE * file = fopen(filename, "w");
	if (file == NULL)
	{
		error("fopen");
		return FALSE;
	}
	fprintf(file, "%u", getpid());
	fclose(file);
	
	return TRUE;
}

void initial_setup()
{
	/* load settings from xml file */
	struct settings * set = load_settings();
	
	/* set global settings */
	changebyinterval = set->changebyinterval;
	interval = set->interval;
	changeeverylaunch = set->changeeverylaunch;
	
	/* list of all wallpapers - no need to lock, no other thread are running */
	wallpapers = g_ptr_array_new();

	struct file_node * node;
	struct profile_node * profile = set->profiles;

	while (profile != NULL)
	{
		if (g_strcmp0(profile->name, set->default_profile) == 0)
		{
			/* copy file nodes */
			struct file_node * node2;
			node = profile->files;
			if (node != NULL)
			{
				files = g_new(struct file_node, 1);
				memcpy(files, node, sizeof(struct file_node));
				files->filename = g_strdup(node->filename);
				node2 = files;
				node = node->next;
				while (node != NULL)
				{
					node2->next = g_new(struct file_node, 1);
					node2 = node2->next;
					memcpy(node2, node, sizeof(struct file_node));
					node2->filename = g_strdup(node->filename);
					node = node->next;
				}
			}
			else
			{
				files = NULL;
			}
		}
		profile = profile->next;
	}
	
	modules = g_ptr_array_new();
	read_modules(set->modules);
	
	free_settings(set);
	
	/* set up history */
	guint i;
	for (i = 0; i < HISTORY_MAX; i++)
	{
		history[i] = 0;
	}
	
	history_current = 0;
	/* TODO: Must this be 0 or 1? */
	history_size = 0;
	
	timer = g_timer_new();
}

int main (int argc, gchar *argv[])
{
	#ifdef USE_DBUS
	g_thread_init(NULL);
	g_type_init();
	dbus_g_thread_init();
	init_dbus();
	#endif
	
	srand(time(0) * getpid());
	
	config_signals();
	
	/* be nice to other processes */
	setpriority(PRIO_PROCESS, 0, 19);
	
	gchar * pid_filename;
	if (! check_pid_file(&pid_filename))
	{
		g_critical("Error in check_pid_file() function. Aborting...");
		return 1;
	}
	
	initial_setup();
	
	if (modules->len > 0)
	{
		scan_files();
		
		if (changebyinterval)
		{
			if (interval < 1)
			{
				changebyinterval = FALSE;
			}
		}
		
		if (wallpapers->len > 1)
		{
			/* main loop - changes wallpaper */
			#ifndef USE_DBUS
			if ((! changeeverylaunch) &&
				(! changebyinterval))
			{
				g_message("Neither Change-By-Interval nor Change-Every-Launch is activated. The program terminates.");
			}
			else
			{
			#endif
			
			if (changeeverylaunch)
				g_timeout_add_seconds(0, &change_launch_timer, NULL);
			
			if (changebyinterval)
				g_timeout_add_seconds(15, &next_wallpaper_timer, NULL);
			
			main_loop = g_main_loop_new(NULL, FALSE);
			if (!quit)
			{
				g_main_loop_run(main_loop);
			}
			g_main_loop_unref(main_loop);
			
			#ifndef USE_DBUS
			}
			#endif
		}
		else
		{
			g_critical("%u wallpapers were found. The program terminates.",
			           wallpapers->len);
		}
		
		clear_wallpapers(wallpapers);
	}
	else
	{
		g_critical("No modules loaded. The program terminates.");
	}
	
	g_timer_destroy(timer);
	
	guint i;
	/* free used modules */
	for (i = 0; i < modules->len; i++)
	{
		struct library_entry * lib = (struct library_entry *)
		                              g_ptr_array_index(modules, i);
		/* mustn't be unloaded, because we use gconf (and glib) with static
		variables in the module! */
		//dlclose(lib->lib);
		g_free(lib->filename);
		g_free(lib);
	}
	g_ptr_array_free(modules, TRUE);
	
	remove(pid_filename);
	g_free(pid_filename);
	
	if (HUP)
	{
		if (execv(argv[0], argv) == -1);
	}
	
	return 0;
}
